/*
 *  file_parser.cc -- BabyTrans ( Babylon Translator front-end for GTK )
 *
 *  Copyright (C) 1999-2002  Frederic Jolliton -- <babytrans@tuxee.net>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <iostream>
#include <algorithm>
#include <string>
#include <vector>
#include <cctype>

using std::isalpha ;
using std::isspace ;

#include <sys/types.h>
#include <regex.h>

#include "file_parser.h"

namespace {

const char			commentDelimiter = '#' ;

bool
isNotBlank( char c )
{
	return ! isspace( c ) ;
}

regex_t				reLine ;
regex_t				reArg ;

void
initRegex()
{
	static bool			initialized = false ;
	if ( initialized ) { return ; }

	regcomp( &reLine ,
		"^\\([[:alpha:]_][[:alpha:][:digit:]_]\\+\\)" // identifier
		"[[:blank:]]*=[[:blank:]]*" // =
		"\\([^[:cntrl:][:space:]=]\\+" // parameters
		"\\([[:blank:]]*,[[:blank:]]*[^[:cntrl:][:space:]=]\\+\\)\\?\\+\\)" // more parameters
		"[[:space:]]*$" , 0 ) ;

	regcomp( &reArg ,
		"\\([^[:cntrl:][:space:]=]\\+\\)" // first parameters
		"[[:blank:]]*\\(,[[:blank:]]*\\)\\?" , // separator
		0 ) ;

	initialized = true ;
}

} // anonymous namespace

/*
 * Synopsis:
 *
 *   bool readConfigLine( std::istream& is , std::vector< std::string >& vs )
 *
 * Description:
 *
 *   Read a single line from a configuration file (via a std::istream)
 *
 *   The syntax is:
 *   identifier = param1 [, param2]*
 *
 *   A '#' character start a comment, and the rest of the line
 *   is simply ignored.
 *
 *   Blank line are ignored.
 *
 * Return value:
 *
 *   Return true if a line has been read without syntax error,
 *   or if the line is empty, and if stream is not in 'bad' or 'fail'
 *   state.
 *   Otherwise, return false.
 */

FileParserStatus
readConfigLine( std::istream& is , std::vector< std::string >& option )
{
	/* ensure regular expression is compiled before using it */
	initRegex() ;

	/* Check stream state */
	if ( is.eof() && ! is.bad() ) {
		/* FIXME: correct ? Is EOF and not BAD mean really end-of-file ? */
		is.setstate( std::ios::failbit ) ;
		return fileParserEof ;
	}

	if ( ! is.good() ) {
		is.setstate( std::ios::failbit ) ;
		return fileParserIoError ;
	}

	std::string			s ;

	/* Read a line from configuration file */
	getline( is , s ) ;

	/* Check stream state */
	if ( is.bad() )
		return fileParserIoError ;

	if ( is.fail() )
		return is.eof()
			? fileParserEof
			: fileParserIoError ;

	/* Delete comment part */
	size_t				pos = s.find( commentDelimiter ) ;
	if ( pos != std::string::npos )
		s.erase( pos ) ;
	
	/* Discard empty line */
	if ( find_if( s.begin() , s.end() , isNotBlank ) == s.end() ) {
		return fileParserOk ;
	}

	/* Use of the regmatch array:
	 *
	 * Index | for reLine:           | for reArg:
	 * ------+-----------------------+------------------
	 * 0     | the complete matched string
	 * 1     | the keyword           | the first argument
	 * 2     | the list of arguments | the separator if any
	 *
	 */
	regmatch_t				r[ 3 ] ;

	if ( regexec( &reLine , s.c_str() , 3 , r , 0 ) != 0 ) {
		return fileParserSyntaxError ;
	}

	std::string				keyword = s.substr( r[ 1 ].rm_so , r[ 1 ].rm_eo ) ;
	std::string				arguments = s.substr( r[ 2 ].rm_so , r[ 2 ].rm_eo ) ;

	option.clear() ;

	/* Add keyword name */
	option.push_back( keyword ) ;

	for ( ; ; ) {
		// NOTE: reuse 'r' array
		if ( regexec( &reArg , arguments.c_str() , 3 , r , 0 ) != 0 ) {
			std::cerr << "boom\n" ;
			return fileParserSyntaxError ;
		}
		/* Add argument */
		option.push_back( arguments.substr( r[ 1 ].rm_so , r[ 1 ].rm_eo ) ) ;

		/* Test if more argument follow (separated by ',') */
		if ( r[ 2 ].rm_so == -1 ) {
			break ;
		}
		/* Keep rest of arguments */
		arguments = arguments.substr( r[ 0 ].rm_eo ) ;
	}

	return fileParserOk ;
}

// vim: set ts=4 sw=4:
